@@ -67,6 +67,10 @@ struct _GnomeFileEntryPrivate {
 	/* FIXME: multiple_entry for entering multiple filenames */
 };
 
+/* Private functions */
+/* Expand files, useful here and in GnomeIconEntry */
+char * _gnome_file_entry_expand_filename (const char *input,
+					  const char *default_dir);
 
 
 static void gnome_file_entry_class_init (GnomeFileEntryClass *class);
@@ -149,22 +153,45 @@ enum {
 
 static int gnome_file_entry_signals[LAST_SIGNAL] = {0};
 
+static gboolean
+gnome_file_entry_mnemonic_activate (GtkWidget *widget,
+				   gboolean   group_cycling)
+{
+	gboolean handled;
+	GnomeFileEntry *entry;
+	
+	entry = GNOME_FILE_ENTRY (widget);
+
+	group_cycling = group_cycling != FALSE;
+
+	if (!GTK_WIDGET_IS_SENSITIVE (entry->_priv->gentry))
+		handled = TRUE;
+	else
+		g_signal_emit_by_name (entry->_priv->gentry, "mnemonic_activate", group_cycling, &handled);
+
+	return handled;
+}
+
 static void
 gnome_file_entry_class_init (GnomeFileEntryClass *class)
 {
 	GtkObjectClass *object_class;
 	GObjectClass *gobject_class;
-
+	GtkWidgetClass *widget_class;
+	
 	parent_class = g_type_class_peek_parent (class);
 
 	object_class = (GtkObjectClass *) class;
 	gobject_class = (GObjectClass *) class;
-
+	widget_class = (GtkWidgetClass *) class;
+	
 	object_class->destroy = gnome_file_entry_destroy;
 
 	gobject_class->finalize = gnome_file_entry_finalize;
 	gobject_class->set_property = fentry_set_property;
 	gobject_class->get_property = fentry_get_property;
+
+	widget_class->mnemonic_activate = gnome_file_entry_mnemonic_activate;
 	
 	gnome_file_entry_signals[BROWSE_CLICKED_SIGNAL] =
 		g_signal_new("browse_clicked",
@@ -385,13 +412,84 @@ browse_dialog_kill (GtkWidget *widget, gpointer data)
 	fentry->fsw = NULL;
 }
 
+/* Does tilde (home directory) expansion on a string */
+static char *
+tilde_expand (const char *str)
+{
+	struct passwd *passwd;
+	const char *p;
+	char *name;
+
+	g_assert (str != NULL);
+
+	if (*str != '~')
+		return g_strdup (str);
+
+	str++;
+
+	p = (const char *)strchr (str, G_DIR_SEPARATOR);
+
+	/* d = "~" or d = "~/" */
+	if (!*str || *str == G_DIR_SEPARATOR) {
+		passwd = getpwuid (geteuid ());
+		p = (*str == G_DIR_SEPARATOR) ? str + 1 : "";
+	} else {
+		if (!p) {
+			p = "";
+			passwd = getpwnam (str);
+		} else {
+			name = g_strndup (str, p - str);
+			passwd = getpwnam (name);
+			g_free (name);
+		}
+	}
+
+	/* If we can't figure out the user name, bail out */
+	if (!passwd)
+		return NULL;
+
+	return g_strconcat (passwd->pw_dir, G_DIR_SEPARATOR_S, p, NULL);
+}
+
+static gchar *
+build_filename (GnomeFileEntry *fentry)
+{
+	const char *text;
+	char *file;
+
+	g_return_val_if_fail (fentry != NULL, NULL);
+	g_return_val_if_fail (GNOME_IS_FILE_ENTRY (fentry), NULL);
+
+	text = gtk_entry_get_text
+		(GTK_ENTRY (gnome_file_entry_gtk_entry (fentry)));
+
+	if (text == NULL || text[0] == '\0')
+		return NULL;
+
+	file = _gnome_file_entry_expand_filename (text, fentry->default_path);
+	if (file == NULL)
+		return NULL;
+
+	/* Now append an '/' if it doesn't exist and we're in directory only mode */
+	if (fentry->_priv->directory_entry && strlen(file) > 0 &&
+	    file[strlen (file)] != G_DIR_SEPARATOR ) {
+		gchar *tmp;
+
+		tmp = g_strconcat (file, "/", NULL);
+
+		g_free (file);
+		return tmp;
+	}
+
+	return file;
+}
+
 static void
 browse_clicked(GnomeFileEntry *fentry)
 {
 	GtkWidget *fsw;
 	GtkFileSelection *fs;
-	GtkWidget *parent;
-	const char *p;
+	char *p;
 
 	/*if it already exists make sure it's shown and raised*/
 	if(fentry->fsw) {
@@ -401,13 +499,11 @@ browse_clicked(GnomeFileEntry *fentry)
 		fs = GTK_FILE_SELECTION(fentry->fsw);
 		gtk_widget_set_sensitive(fs->file_list,
 					 ! fentry->_priv->directory_entry);
-		p = gtk_entry_get_text (GTK_ENTRY (gnome_file_entry_gtk_entry (fentry)));
-		if(p && *p!=G_DIR_SEPARATOR && fentry->default_path) {
-			char *dp = g_build_filename (fentry->default_path, p, NULL);
-			gtk_file_selection_set_filename (fs, dp);
-			g_free(dp);
-		} else {
+
+		p = build_filename (fentry);
+		if (p != NULL) {
 			gtk_file_selection_set_filename (fs, p);
+			g_free (p);
 		}
 		return;
 	}
@@ -416,37 +512,6 @@ browse_clicked(GnomeFileEntry *fentry)
 	fsw = gtk_file_selection_new (fentry->_priv->browse_dialog_title
 				      ? fentry->_priv->browse_dialog_title
 				      : _("Select file"));
-	/* BEGIN UGLINESS.  This code is stolen from gnome_dialog_set_parent.
-	 * We want its functionality, but it takes a GnomeDialog as its argument.
-	 * So we copy it )-: */
-	parent = gtk_widget_get_toplevel (GTK_WIDGET (fentry));
-	gtk_window_set_transient_for (GTK_WINDOW(fsw), GTK_WINDOW (parent));
-
-	/* FIXME! */
-	if ( 1 ) { /*gnome_preferences_get_dialog_centered() ) { */
-
-		/* User wants us to center over parent */
-
-		gint x, y, w, h, dialog_x, dialog_y;
-
-		if ( ! GTK_WIDGET_VISIBLE(parent)) return; /* Can't get its
-							      size/pos */
-
-		/* Throw out other positioning */
-		gtk_window_set_position(GTK_WINDOW(fsw),GTK_WIN_POS_NONE);
-
-		gdk_window_get_origin (GTK_WIDGET(parent)->window, &x, &y);
-		gdk_drawable_get_size   (GTK_WIDGET(parent)->window, &w, &h);
-
-		/* The problem here is we don't know how big the dialog is.
-		   So "centered" isn't really true. We'll go with
-		   "kind of more or less on top" */
-
-		dialog_x = x + w/4;
-		dialog_y = y + h/4;
-
-		gtk_window_move (GTK_WINDOW (fsw), dialog_x, dialog_y);
-	}
 
 	g_object_set_data (G_OBJECT (fsw), "gnome_file_entry", fentry);
 
@@ -454,14 +519,9 @@ browse_clicked(GnomeFileEntry *fentry)
 	gtk_widget_set_sensitive(fs->file_list,
 				 ! fentry->_priv->directory_entry);
 
-	p = gtk_entry_get_text (GTK_ENTRY (gnome_file_entry_gtk_entry (fentry)));
-	if(p && *p!=G_DIR_SEPARATOR && fentry->default_path) {
-		char *dp = g_build_filename (fentry->default_path, p, NULL);
-		gtk_file_selection_set_filename (fs, dp);
-		g_free(dp);
-	} else {
-		gtk_file_selection_set_filename (fs, p);
-	}
+	p = build_filename (fentry);
+	gtk_file_selection_set_filename (fs, p);
+	g_free (p);
 
 	g_signal_connect (fs->ok_button, "clicked",
 			  G_CALLBACK (browse_dialog_ok),
@@ -589,7 +649,7 @@ gnome_file_entry_init (GnomeFileEntry *fentry)
 	gtk_box_pack_start (GTK_BOX (hbox), fentry->_priv->gentry, TRUE, TRUE, 0);
 	gtk_widget_show (fentry->_priv->gentry);
 
-	button = gtk_button_new_with_label (_("Browse..."));
+	button = gtk_button_new_with_mnemonic (_("_Browse..."));
 	g_signal_connect (button, "clicked",
 			  G_CALLBACK (browse_clicked_signal),
 			  fentry);
@@ -767,43 +827,32 @@ gnome_file_entry_set_default_path(GnomeFileEntry *fentry, const char *path)
 	fentry->default_path = p;
 }
 
-/* Does tilde (home directory) expansion on a string */
-static char *
-tilde_expand (char *str)
+/* Expand files, useful here and in GnomeIconEntry */
+char *
+_gnome_file_entry_expand_filename (const char *input, const char *default_dir)
 {
-	struct passwd *passwd;
-	char *p;
-	char *name;
-
-	g_assert (str != NULL);
-
-	if (*str != '~')
-		return g_strdup (str);
-
-	str++;
-
-	p = strchr (str, G_DIR_SEPARATOR);
-
-	/* d = "~" or d = "~/" */
-	if (!*str || *str == G_DIR_SEPARATOR) {
-		passwd = getpwuid (geteuid ());
-		p = (*str == G_DIR_SEPARATOR) ? str + 1 : "";
-	} else {
-		if (!p) {
-			p = "";
-			passwd = getpwnam (str);
-		} else {
-			name = g_strndup (str, p - str);
-			passwd = getpwnam (name);
-			g_free (name);
-		}
-	}
-
-	/* If we can't figure out the user name, bail out */
-	if (!passwd)
+	if (input == NULL) {
 		return NULL;
-
-	return g_strconcat (passwd->pw_dir, G_DIR_SEPARATOR_S, p, NULL);
+	} else if (input[0] == '\0') {
+		return g_strdup ("");
+	} else if (input[0] == G_DIR_SEPARATOR) {
+		return g_strdup (input);
+	} else if (input[0] == '~') {
+		return tilde_expand (input);
+	} else if (default_dir != NULL) {
+		char *ret = g_build_filename (default_dir, input, NULL);
+		if (ret[0] == '~') {
+			char *tmp = tilde_expand (ret);
+			g_free (ret);
+			return tmp;
+		}
+		return ret;
+	} else {
+		char *cwd = g_get_current_dir ();
+		char *ret = g_build_filename (cwd, input, NULL);
+		g_free (cwd);
+		return ret;
+	}
 }
 
 /**
@@ -825,67 +874,45 @@ tilde_expand (char *str)
 char *
 gnome_file_entry_get_full_path(GnomeFileEntry *fentry, gboolean file_must_exist)
 {
-	char *p;
-	char *t;
+	const char *text;
+	char *file;
 
 	g_return_val_if_fail (fentry != NULL, NULL);
 	g_return_val_if_fail (GNOME_IS_FILE_ENTRY (fentry), NULL);
 
-	t = gtk_editable_get_chars (GTK_EDITABLE (gnome_file_entry_gtk_entry (fentry)), 0, -1);
+	text = gtk_entry_get_text
+		(GTK_ENTRY (gnome_file_entry_gtk_entry (fentry)));
 
-	if (!t)
+	if (text == NULL || text[0] == '\0')
 		return NULL;
-	else if (!*t) {
-		g_free (t);
+
+	file = _gnome_file_entry_expand_filename (text, fentry->default_path);
+	if (file == NULL)
 		return NULL;
-	}
-
-	if (*t == G_DIR_SEPARATOR)
-		p = t;
-	else if (*t == '~') {
-		p = tilde_expand (t);
-		g_free (t);
-	} else if (fentry->default_path) {
-		p = g_build_filename (fentry->default_path, t, NULL);
-		g_free (t);
-		if (*p == '~') {
-			t = p;
-			p = tilde_expand (t);
-			g_free (t);
-		}
-	} else {
-		gchar *cwd = g_get_current_dir ();
-
-		p = g_build_filename (cwd, t, NULL);
-		g_free (cwd);
-		g_free (t);
-	}
 
 	if (file_must_exist) {
-		if (!p)
-			return NULL;
-
 		if (fentry->_priv->directory_entry) {
 			char *d;
 
-			if (g_file_test (p, G_FILE_TEST_IS_DIR))
-				return p;
+			if (g_file_test (file, G_FILE_TEST_IS_DIR))
+				return file;
 
-			d = g_path_get_dirname (p);
-			g_free (p);
+			d = g_path_get_dirname (file);
+			g_free (file);
 
 			if (g_file_test (d, G_FILE_TEST_IS_DIR))
 				return d;
+			g_free (d);
 
-			p = d;
-		} else if (g_file_test (p, G_FILE_TEST_EXISTS)) {
-			return p;
+			return NULL;
+		} else if (g_file_test (file, G_FILE_TEST_EXISTS)) {
+			return file;
 		}
 
-		g_free (p);
+		g_free (file);
 		return NULL;
 	} else {
-		return p;
+		return file;
 	}
 }
 
