@@ -111,7 +111,7 @@ typedef struct {
 
 	/* User data and destroy notify function */
 	gpointer data;
-	GtkDestroyNotify destroy;
+	GDestroyNotify destroy;
 
 	/* ID for the text item's event signal handler */
 	guint text_event_id;
@@ -1295,6 +1295,9 @@ gnome_icon_list_remove (GnomeIconList *gil, int pos)
 	if (priv->icons >= priv->last_selected_idx)
 		priv->last_selected_idx = -1;
 
+	if (priv->icons >= priv->focus_icon)
+		priv->focus_icon = -1;
+	
 	if (priv->last_selected_icon == icon)
 		priv->last_selected_icon = NULL;
 
@@ -1598,6 +1601,7 @@ real_move_cursor (Gil *gil, GtkDirectionType dir, gboolean clear_selection)
 			return;
 		break;
 	default:
+		break;
 	}
 
 	if (clear_selection) {
@@ -2545,7 +2549,7 @@ gnome_icon_list_set_selection_mode (GnomeIconList *gil, GtkSelectionMode mode)
 void
 gnome_icon_list_set_icon_data_full (GnomeIconList *gil,
 				    int pos, gpointer data,
-				    GtkDestroyNotify destroy)
+				    GDestroyNotify destroy)
 {
 	Icon *icon;
 
